pub fn blockDataTransfer(comptime InstrFn: type, comptime P: bool, comptime U: bool, comptime S: bool, comptime W: bool, comptime L: bool) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        fn inner(cpu: *Arm32, opcode: u32) void {
            const rn: u4 = @truncate(opcode >> 16 & 0xF);
            const rlist: u16 = @intCast(opcode & 0xFFFF);

            const reg_count: u32 = @popCount(rlist);
            const first_in_list: u4 = @truncate(@ctz(rlist)); // note that @ctz(0x0000) and @ctz(0x0001) collide

            // U determines whether the LDM/STM transfer is made upwards (U == 1)
            // or downwards (U == 0).

            const base_addr = cpu.r[rn];

            const start_addr: u32 = if (U) blk: {
                break :blk base_addr + if (P) 4 else 0;
            } else blk: {
                break :blk base_addr - (4 * reg_count) + if (!P) 4 else 0;
            };

            const new_base_addr: u32 = if (U) blk: {
                break :blk base_addr + 4 * reg_count;
            } else blk: {
                break :blk base_addr - 4 * reg_count;
            };

            var address = start_addr;

            // On Empty List:
            //
            // ARMv4 Only: R15 is Loaded/Stored
            // ARMv4/Armv5: Rn = Rn +/- 0x40

            if (rlist == 0) {
                if (Arm32.arch == .v4t) {
                    const undefined_addr: u32 = if (U) blk: {
                        break :blk base_addr + if (P) 4 else 0;
                    } else blk: {
                        break :blk base_addr - (0x40 - if (!P) 4 else 0);
                    };

                    if (L) {
                        cpu.r[15] = cpu.read(u32, undefined_addr);
                        cpu.pipe.reload(cpu);
                    } else {
                        cpu.write(u32, undefined_addr, cpu.r[15] + 4);
                    }
                }

                cpu.r[rn] = if (U) base_addr + 0x40 else base_addr - 0x40;
                return;
            }

            for (first_in_list..16) |idx| {
                const i: u4 = @intCast(idx);

                if (rlist >> i & 1 == 1) {
                    if (L) {
                        load(cpu, i, rlist, address);
                    } else {
                        store(cpu, rn, i, rlist, address, .{ .old_addr = base_addr, .new_addr = new_base_addr });
                    }

                    address += 4;
                }
            }

            if (W and !L)
                cpu.r[rn] = new_base_addr;

            if (W and L) {
                // What happens when W is set and Rn is in the rlist? (LDM)
                //
                // ARMv4: No writeback
                // ARMv5: writeback if Rn is "the ONLY register" or NOT the LAST register

                if (rlist >> rn & 1 == 0) { // rn is not in rlist
                    cpu.r[rn] = new_base_addr;
                    return;
                }

                switch (Arm32.arch) {
                    .v4t => {}, // No Writeback
                    .v5te => {
                        const rn_is_last = (15 - @clz(rlist)) <= rn;

                        if (reg_count == 1 or !rn_is_last) {
                            cpu.r[rn] = new_base_addr;
                        }
                    },
                }
            }
        }

        fn load(cpu: *Arm32, ri: u4, rlist: u16, address: u32) void {
            const has_r15 = rlist >> 15 & 1 == 1;

            if (S and !has_r15) {
                // Always Transfer User mode Registers
                cpu.setUserModeRegister(ri, cpu.read(u32, address));
            } else {
                const value = cpu.read(u32, address);
                cpu.r[ri] = value;

                if (ri == 0xF) {
                    const mask: u32 = if (Arm32.arch == .v5te) 1 else 3;
                    cpu.r[ri] &= ~mask;

                    if (Arm32.arch == .v5te) cpu.cpsr.t.write(value & 1 == 1);
                    if (S) cpu.setCpsr(cpu.spsr.raw); // FIXME: before or after the reload?

                    cpu.pipe.reload(cpu);
                }
            }
        }

        const BaseAddrs = struct { old_addr: u32, new_addr: u32 };

        fn store(cpu: *Arm32, rn: u4, ri: u4, rlist: u16, address: u32, base: BaseAddrs) void {
            const value = if (S) blk: {
                // if S == true:
                //     Always Transfer User mode Registers
                //     This happens regardless if r15 is in the list

                break :blk cpu.getUserModeRegister(ri);
            } else blk: {
                if (ri == rn) {
                    // What happens when W is set and Rn is in the rlist? (STM)
                    //
                    // Armv4: Store OLD Base if Rb is FIRST entry in Rlist, otherwise store NEW base
                    // Armv5: Always store OLD Base

                    if (rlist >> rn & 1 == 0)
                        break :blk base.new_addr;

                    const mask = @as(u16, 1) << rn;
                    const is_first = @popCount(rlist & (mask - 1)) == 0;

                    break :blk switch (Arm32.arch) {
                        .v4t => if (is_first) base.old_addr else base.new_addr,
                        .v5te => base.old_addr,
                    };
                }

                break :blk cpu.r[ri];
            };

            cpu.write(u32, address, value + if (ri == 0xF) 4 else @as(u32, 0));
        }
    }.inner;
}
